<?php

namespace Drupal\fluent;

use Drupal\Core\Entity\EntityBase;
use Drupal\Core\Field\FieldItemInterface;
use Drupal\Core\Field\FieldItemList;
use Drupal\Core\Field\FieldItemListInterface;
use Drupal\Core\Field\Plugin\Field\FieldType\EntityReferenceItem;
use Drupal\field\Entity\FieldConfig;
use Drupal\file\Entity\File;
use Drupal\file\Plugin\Field\FieldType\FileUriItem;
use Drupal\fluent\Plugin\FluentFieldResolverManager;
use Drupal\node\Entity\Node;

/**
 *
 */
class Fluent {


  /**
   * @var \Drupal\fluent\Plugin\FluentFieldResolverManager
   */
  private $pluginManager;

  /**
   * @var array
   */
  private $fieldResolvers;

  /**
   * @var \Drupal\fluent\Plugin\FluentFieldResolver\CommonField
   */
  private $commonResolver;

  /**
   * @var mixed
   */
  private $entity;

  /**
   *
   */
  public function __construct(FluentFieldResolverManager $plugin) {
    $this->pluginManager = $plugin;

    $this->fieldResolvers = [];
    foreach ($this->pluginManager->getDefinitions() as $id => $definition) {
      $this->fieldResolvers[$id] = $this->pluginManager->createInstance($id);
    }

    $this->commonResolver = $this->fieldResolvers['common_field'];
    unset($this->fieldResolvers['common_field']);
  }

  /**
   * @param $entity
   * @return $this
   */
  public function using($entity) {
    $this->entity = $entity;

    return $this;
  }

  public function only(array $paths) {

  }

  /**
   * @param string $path
   * @param $defaultValue
   * @return mixed
   *
   * @throws \Exception
   */
  public function firstValue(string $path, $defaultValue = NULL) {
    return $this->nthValue($path, 0, $defaultValue)->first();
  }

  public function values(string $path, array $default = []): Collection {
    return $this->nthValue($path, -1, $default);
  }

  /**
   *
   */
  public function nthValue(string $path, int $position, $defaultValue = NULL): Collection {
    $path = $this->getPath($path);

    // Save the current entity.
    $tmpEntity = $this->entity;
    $items = count($path);
    $count = 0;

    do {
      $possibleField = $path[$count];

      $solvedField = $this->getResolver($tmpEntity, $possibleField, $position);

      if ($solvedField === NULL) {
        throw new \Exception('Field not found');
      }

      $tmpEntity = $solvedField;
      if ($this->skipCount($solvedField)) {
        continue;
      }

      $count++;
    } while ($count < $items);

    // TODO: Create plugin for each @FieldType
    if ($tmpEntity instanceof FieldItemInterface) {
      $field = $tmpEntity->getString();
    } else {
      $field = $tmpEntity;
    }

    if ($field instanceof EntityBase) {
      $collection = [$field];
    } else {
      $collection = $field;
    }

    return Collection::make($collection);
  }

  private function skipCount($solvedField): bool {
    return (
      $solvedField instanceof FieldItemListInterface
      ||
      $solvedField instanceof EntityReferenceItem
    );
  }

  /**
   * @param string $path
   *
   * @return array
   */
  private function getPath(string $path): array {
    return explode('.', $path);
  }

  /**
   *
   */
  private function getResolver($entity, $name, $position = 0) {
    if (method_exists($entity, 'getFieldDefinition')) {
      /** @var FieldConfig $fieldDefinition */
      $fieldDefinition = $entity->getFieldDefinition($name);
      if ($fieldDefinition) {
        foreach ($this->fieldResolvers as $resolver) {
          if ($resolver->can($fieldDefinition)) {
            $cardinality = $fieldDefinition->getFieldStorageDefinition()->getCardinality();

            return $resolver->handler($entity, $name, [
              'position' => $position,
              'field_cardinality' => $cardinality,
            ]);
          }
        }
      }

      return $this->commonResolver->handler($entity, $name);
    } else {
      // TODO: don't fieldable entity
    }
  }

  private function fieldItemResolver(FieldItemInterface $fieldItem) {
  }
}
